;;; coding-cookie.el --- processing coding cookies in files like in emacs

;; Copyright (C) 1995 Electrotechnical Laboratory, JAPAN.
;; Copyright (C) 2001 Free Software Foundation, Inc.
;; Licensed to the Free Software Foundation.

;; Keywords: mule, multilingual, character set, coding system

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.


;; All code is from emacs, except `run-auto-coding-function', which
;; ensures calling emacs code like in emacs.  Changes of code from
;; emacs are minimal and include:

;; adapting for coding-system-p which does not do checks for symbols;

;; adding save-match-data;

;; removing auto-coding-alist both from code and documentation;

;; removing auto-coding-alist-lookup;

;; adding autoload cookies.


;;; start of code from emacs 21.1
;;;###autoload
(defcustom auto-coding-regexp-alist
  '(("^BABYL OPTIONS:[ \t]*-\\*-[ \t]*rmail[ \t]*-\\*-" . no-conversion))
  "Alist of patterns vs corresponding coding systems.
Each element looks like (REGEXP . CODING-SYSTEM).
A file whose first bytes match REGEXP is decoded by CODING-SYSTEM on reading.

The settings in this alist take priority over `coding:' tags
in the file (see the function `set-auto-coding')
and the contents of `file-coding-system-alist'."
  :group 'files
  :group 'mule
  :type '(repeat (cons (regexp :tag "Regexp")
		       (symbol :tag "Coding system"))))

(defvar set-auto-coding-for-load nil
  "Non-nil means look for `load-coding' property instead of `coding'.
This is used for loading and byte-compiling Emacs Lisp files.")


(defun coding-system-or-symbol-p (coding-system)
  "Similar to emacs coding-system-p.  Returns t both for coding system
object and for symbol that is a coding system name."
  (or (coding-system-p coding-system)
      (coding-system-p (find-coding-system coding-system))))


(defun auto-coding-from-file-contents (size)
  "Determine a coding system from the contents of the current buffer.
The current buffer contains SIZE bytes starting at point.
Value is either a coding system or nil."
  (save-excursion
    (let ((alist auto-coding-regexp-alist)
	  coding-system)
      (while (and alist (not coding-system))
	(let ((regexp (car (car alist))))
	  (when (re-search-forward regexp (+ (point) size) t)
	    (setq coding-system (cdr (car alist)))))
	(setq alist (cdr alist)))
      coding-system)))


;;;###autoload
(defun set-auto-coding (filename size)
  "Return coding system for a file FILENAME of which SIZE bytes follow point.
These bytes should include at least the first 1k of the file
and the last 3k of the file, but the middle may be omitted.

It checks the
contents of the current buffer following point against
`auto-coding-regexp-alist'.  If no match is found, it checks for a
`coding:' tag in the first one or two lines following point.  If no
`coding:' tag is found, it checks for local variables list in the last
3K bytes out of the SIZE bytes.

The return value is the specified coding system,
or nil if nothing specified.

The variable `set-auto-coding-function' (which see) is set to this
function by default."
  (save-match-data
  (or (auto-coding-from-file-contents size)
      (let* ((case-fold-search t)
	     (head-start (point))
	     (head-end (+ head-start (min size 1024)))
	     (tail-start (+ head-start (max (- size 3072) 0)))
	     (tail-end (+ head-start size))
	     coding-system head-found tail-found pos)
	;; Try a short cut by searching for the string "coding:"
	;; and for "unibyte:" at the head and tail of SIZE bytes.
	(setq head-found (or (search-forward "coding:" head-end t)
			     (search-forward "unibyte:" head-end t)))
	(if (and head-found (> head-found tail-start))
	    ;; Head and tail are overlapped.
	    (setq tail-found head-found)
	  (goto-char tail-start)
	  (setq tail-found (or (search-forward "coding:" tail-end t)
			       (search-forward "unibyte:" tail-end t))))

	;; At first check the head.
	(when head-found
	  (goto-char head-start)
	  (setq pos (re-search-forward "[\n\r]" head-end t))
	  (if (and pos
		   (= (char-after head-start) ?#)
		   (= (char-after (1+ head-start)) ?!))
	     ;; If the file begins with "#!" (exec interpreter magic),
	  ;; look for coding frobs in the first two lines.  You cannot
	      ;; necessarily put them in the first line of such a file
	      ;; without screwing up the interpreter invocation.
	      (setq pos (search-forward "\n" head-end t)))
	  (if pos (setq head-end pos))
	  (when (< head-found head-end)
	    (goto-char head-start)
	    (when (and set-auto-coding-for-load
		       (re-search-forward
			"-\\*-\\(.*;\\)?[ \t]*unibyte:[ \t]*\\([^ ;]+\\)"
			head-end t))
	      (setq coding-system 'raw-text))
	    (when (and (not coding-system)
		       (re-search-forward
			"-\\*-\\(.*;\\)?[ \t]*coding:[ \t]*\\([^ ;]+\\)"
			head-end t))
	      (setq coding-system (intern (match-string 2)))
	      (or (coding-system-or-symbol-p coding-system)
		  (setq coding-system nil)))))

	;; If no coding: tag in the head, check the tail.
	(when (and tail-found (not coding-system))
	  (goto-char tail-start)
	  (search-forward "\n\^L" nil t)
	  (if (re-search-forward
	       "^\\(.*\\)[ \t]*Local Variables:[ \t]*\\(.*\\)$" tail-end t)
	  ;; The prefix is what comes before "local variables:" in its
	   ;; line.  The suffix is what comes after "local variables:"
	      ;; in its line.
	      (let* ((prefix (regexp-quote (match-string 1)))
		     (suffix (regexp-quote (match-string 2)))
		     (re-coding
		      (concat
		       "^" prefix
		       ;; N.B. without the \n below, the regexp can
		       ;; eat newlines.
		       "[ \t]*coding[ \t]*:[ \t]*\\([^ \t\n]+\\)[ \t]*"
		       suffix "$"))
		     (re-unibyte
		      (concat
		       "^" prefix
		       "[ \t]*unibyte[ \t]*:[ \t]*\\([^ \t\n]+\\)[ \t]*"
		       suffix "$"))
		     (re-end
		      (concat "^" prefix "[ \t]*End *:[ \t]*" suffix "$"))
		     (pos (point)))
		(re-search-forward re-end tail-end 'move)
		(setq tail-end (point))
		(goto-char pos)
		(when (and set-auto-coding-for-load
			   (re-search-forward re-unibyte tail-end t))
		  (setq coding-system 'raw-text))
		(when (and (not coding-system)
			   (re-search-forward re-coding tail-end t))
		  (setq coding-system (intern (match-string 1)))
		  (or (coding-system-or-symbol-p coding-system)
		      (setq coding-system nil))))))
	coding-system)))
  )
;;; end of code from emacs 21.1

;;; the following is implemented in emacs in C
(defvar set-auto-coding-function 'set-auto-coding
  "If non-nil, a function to call to decide a coding system of file.
Two arguments are passed to this function: the file name
and the length of a file contents following the point.
This function should return a coding system to decode the file contents.
If no coding system is decided, it should check a coding system
specified in the heading lines with the format:
	-*- ... coding: CODING-SYSTEM; ... -*-
or local variable spec of the tailing lines with `coding:' tag.")

;; emacs C code does similar things unconditionally
;;;###autoload
(defun run-auto-coding-function (filename visit)
  (and
   (boundp 'set-auto-coding-function) set-auto-coding-function
   (let ((tn (file-truename filename)))
     (when (file-regular-p tn)
       (let ((size (nth 7 (file-attributes tn))))
	 (save-excursion
	   (set-buffer (get-buffer-create " *code-converting-work*"))
	   (erase-buffer)
	   (let* ((hs 1024) (ts 3072) (ss (+ hs ts)))
	     (if (<= size ss)
		 (insert-file-contents-internal
		  filename nil nil nil nil 'binary)
	       (insert-file-contents-internal
		filename nil 0 hs nil 'binary)
	       (goto-char (point-max))
	       (insert-file-contents-internal
		filename nil (- size ts)  size nil 'binary)
	       (setq size ss)
	       (goto-char (point-min))))
	   (funcall set-auto-coding-function filename size)))))))

;;;###autoload
(defun coding-cookie-enable ()
  "Enables Emacs-ish methods of recognizing coding of files being read
in."
  (interactive)
  (add-hook
   'insert-file-contents-pre-hook
   'run-auto-coding-function))

;;;###autoload
(defun coding-cookie-disable ()
  "Disables Emacs-ish methods of recognizing coding of files being
read in."
  (interactive)
  (remove-hook
   'insert-file-contents-pre-hook
   'run-auto-coding-function))

(provide 'coding-cookie)
;;; coding-cookie.el ends here
